import os
import csv
from rdkit import Chem
from rdkit.Chem import AllChem, Descriptors
import openbabel as ob
import subprocess
import sys

def osCommand(command):
    c  = subprocess.Popen(command, shell=True)
    try:
        c.wait(120)
    except subprocess.TimeoutExpired:
        print("Command timed out")
        c.kill()
        pass
    except Exception:
        pass

def get_energy(file_name):
    file = open(file_name) 
    lines = file.readlines() 
    file.close() 
    line = lines[1] 
    result = float(line.split(':')[1].split()[0])  #splits between : and 0, the binding energy is always between these on line[1]
    return result

def process(mol, receptors, l_num, name):
    print("Trying to acquire ligand attributes and binding energy...")
    temp_array = []
    recept_num = 1
    temp_array.append('ligand_'+str(l_num))
    temp_array.append(Descriptors.MolWt(mol))
    temp_array.append(Descriptors.MolLogP(mol))
    temp_array.append(Chem.MolToSmiles(mol))
    if mol.HasProp('NSC'):
        temp_array.append(mol.GetProp('NSC'))
    elif mol.HasProp('E_NSC'):
        temp_array.append(mol.GetProp('E_NSC'))
    else:
        temp_array.append('N/A')
    for receptor in receptors:
        ligand_file = os.path.join('data', str(name), 'vina_files', 'ligand_'+str(l_num)+'-'+receptor+'.pdbqt')
        temp_array.append(str(get_energy(ligand_file)))
    csv_fname = str(name)+"_results.csv"
    csv_file = os.path.join('data', name, csv_fname)
    with open(csv_file, "a") as f:
        writer = csv.writer(f, delimiter=",")
        writer.writerow(temp_array)
    print("Ligand attributes and binding energy acquired and appended to csv file")

def makedir(dir):
    if not os.path.isdir(dir):
        os.makedirs(dir)

def convertMol(pdb, pdbqt):
    obCon = ob.OBConversion()
    obCon.SetInAndOutFormats('pdb', 'pdbqt')
    mol = ob.OBMol()
    obCon.ReadFile(mol, pdb)
    obCon.WriteFile(mol, pdbqt)
    os.remove(pdb)

def docking(pdbqt_file, receptors, vina_files_dir, molnum):
    recept_num = 0
    for receptor in receptors:
        # run vina
        print("Docking ligand " + str(molnum) + " with " + receptor)
        f_out_pdbqt = os.path.join(vina_files_dir, 'ligand_' + str(molnum) + '-' + str(receptor) + '.pdbqt')
        f_out_log = os.path.join(vina_files_dir, 'ligand_' + str(molnum) + '-' + str(receptor) + '.txt')
        # python = sys.executable.split(os.sep)
        # vina = os.path.join(os.sep, *python[:-1], 'vina')
        rpath = os.path.join('receptors', receptor)
        vina_command = 'vina --config ' + rpath + '-config.txt --ligand ' + pdbqt_file + ' --out ' + f_out_pdbqt + ' --log ' + f_out_log
        print(vina_command)
        osCommand(vina_command)
        recept_num += 1

def moldocking(name, start_ligand, receptors):
    data_dir = os.path.join('data', name)
    vina_ligands_dir = os.path.join(data_dir, 'vina_ligands')
    makedir(vina_ligands_dir)
    vina_files_dir = os.path.join(data_dir, 'vina_files')
    makedir(vina_files_dir)
    #start docking:
    suppl = Chem.SDMolSupplier(os.path.join(data_dir, str(name)+'.sdf'))
    pdb_file = os.path.join(data_dir, 'ligand.pdb')
    pdbqt_file = pdb_file + "qt"
    molnum = 1
    for mol in suppl:
        if molnum >= start_ligand:
            AllChem.EmbedMolecule(mol)
            AllChem.UFFOptimizeMolecule(mol)
            AllChem.MolToPDBFile(mol, pdb_file)
            convertMol(pdb_file, pdbqt_file)
            docking(pdbqt_file, receptors, vina_files_dir, molnum)
            os.remove(pdbqt_file)
            process(mol, receptors, molnum, name)
        else:
            pass
        molnum += 1



from rdkit import Chem, DataStructs
from rdkit.ML.Cluster import Butina
from rdkit.Chem import Draw, AllChem
import matplotlib.pyplot as plt
import numpy as np
import csv
import os

def ClusterFps(fps,cutoff):
    # first generate the distance matrix:
    dists = []
    nfps = len(fps)
    for i in range(1,nfps):
        sims = DataStructs.BulkTanimotoSimilarity(fps[i],fps[:i])
        dists.extend([1-x for x in sims])
    # now cluster the data:
    cs = Butina.ClusterData(dists,nfps,cutoff,isDistData=True, reordering=True)
    return cs

def StandardDev(avg, values, length):
    sum_sqrs = 0
    for val in values:
        sqr = (val - avg) ** 2
        sum_sqrs += sqr
    mean_sqrs = sum_sqrs / length
    sd = mean_sqrs ** 0.5
    return sd


#def learn(input):
learn_xp = input("Enter data set to learn:")
input_learn_csv = os.path.join('data', learn_xp, learn_xp+'_results.csv')

infile = open(input_learn_csv, "r")
reader = csv.reader(infile, delimiter=',')
rownum = 0
mols = []
print('Extracting data...')
for row in reader:
    if rownum == 0:
        pass
    else:
        if row[5] == 'N/A':
            pass
        else:
            np_fp = np.array(0,)
            mol = Chem.MolFromSmiles(row[3])
            mol.SetProp('Energy R1', row[5])
            mol.SetProp('Energy R2', row[6])
            mol.SetProp('NSC', row[4])
            #set other props here for clustering e.g. TO displacement, etc.
            mols.append(mol)
    rownum += 1
infile.close()
print('Done.')

print('Creating clusters...')
fps = [AllChem.GetMorganFingerprintAsBitVect(mol, 2, 1024) for mol in mols]
clusters = np.array(ClusterFps(fps, 0.5))
print('Done.')

cnt = 0
avg_nrgs = []
sds = []
clustered = 0
clstd_mols = []
for cluster in clusters:
    ncluster = len(cluster)
    sum_nrgs = 0
    nrgs = []
    temp_mols = []
    for i in cluster:
        nrg = float(mols[i].GetProp('Energy R1')) #replace
        sum_nrgs += nrg
        nrgs.append(nrg)
        temp_mols.append(mols[i])
    avg = sum_nrgs / ncluster
    sd = StandardDev(avg, nrgs, ncluster)
    # print('Average for cluster ' + str(cnt) + ': ' + str(avg) + ' p/m ' + str(sd) + '. nmols = ' + str(ncluster))
    if sd > 0:
        clstd_mols.append(temp_mols)
        avg_nrgs.append(avg)
        sds.append(sd)
        clustered += ncluster
    cnt += 1
print('Total clusters: ' + str(cnt))
print('Total compounds in clusters > 1: ' + str(clustered))
print('Best cluster is ' + str(avg_nrgs.index(min(avg_nrgs))) + ' with energy of ' + str(min(avg_nrgs)))

best = clstd_mols[avg_nrgs.index(min(avg_nrgs))]
nbest = len(best)
img = Draw.MolsToGridImage(best[:nbest], molsPerRow=3, legends=[str(i.GetProp('NSC') + ', ' + i.GetProp('Energy R1')) for i in best]) #replace
img.save('best-mols-by-energy-R1.png') #replace

plt.figure()
x = range(0, len(avg_nrgs))
plt.xlabel('Cluster Number')
plt.ylabel('Average Binding Energy (kcal / mol) of pocket 1') #replace
plt.errorbar(x, avg_nrgs, yerr=sds, fmt='s', capsize=3)
plt.savefig('clusters-by-energy-R1.png') #replace
print('Figure saved.')




import os
from rdkit import Chem
from rdkit.Chem import AllChem, Descriptors
import random

def findLargestRingNum(s):
    max_num = 0
    for x in range(len(s)):
        if s[x].isdigit() and int(s[x]) > max_num:
            max_num = int(s[x])
    return max_num

def replace_R(mol, fg):
    wc = Chem.MolFromSmiles("[*]")
    newmol = Chem.ReplaceSubstructs(mol, wc, fg)
    Chem.SanitizeMol(newmol[0])
    return newmol[0]

def getAverages(list):
    tot = 0
    num = 0
    for i in list:
        tempmol = Chem.MolFromSmiles(i)
        tot += Descriptors.MolWt(tempmol)
        num += 1
    avg = tot / num
    return avg

def writeMol(mol, sdf_path):
    writer = Chem.SDWriter(sdf_path)
    writer.write(mol)

#MOLECULE GENERATION
def molgenerate(name, target_mass, nligands, mol, s_list, nt_list, t_list):
    if mol is not None:
        smi = Chem.MolToSmiles(mol)
        wc = Chem.MolFromSmiles('[*]')
        wcFactor = len(mol.GetSubstructMatches(wc))
        s_list = [smi]
    else:
        wcFactor = 1
    avgSGp = getAverages(s_list)  # gets avg of the masses of each dictionary
    avgNTGp = getAverages(nt_list)
    avgTGp = getAverages(t_list)
    app_mass = target_mass - avgSGp - (avgTGp * wcFactor)  # subs avg mass of s and t groups from target
    NTGps_to_use = int((app_mass // (avgNTGp * wcFactor)) * wcFactor) # finds out how many nt groups, '//' = rounded down to nearest integer
    # if this rounds to 0  then no nt groups will be used just s and t groups!
    perm_max = [len(s_list)] + ([len(nt_list)] * NTGps_to_use) + ([len(t_list)] * wcFactor) #this is the max num of each group that will be gen'd
    perm_length = len(perm_max)
    if nligands > 0:
        print('Generating ' + str(nligands) + ' compounds. This can take a while...')
    else:
        print("Generating all compounds. This can take a while...")
    perms = []
    if nligands > 0: #if the user only wants a select number of molecules then...
        while len(perms) < nligands:
            #check the length of the permutations vs number of ligands wanted
            perm = []
            pos = 0
            while pos < perm_length:
                #for each position in a permutation, generate a random number
                perm.append(random.choice(range(perm_max[pos])))
                pos += 1
            found = perm in perms
            #check if this permutation exists, if it does, try again
            if found is True:
                pass
            else:
                perms.append(perm)
    else:
        total_cmpds = 1
        for i in perm_max:
            total_cmpds = total_cmpds * i  # calculates the total permutations i.e. total cmpds to gen
        cont = input("This will generate " + str(total_cmpds) + " compounds...\nDo you wish to continue? (y/n) ")
        if cont == 'y' or cont == 'yes':
            pass
        else:
            return()
        pos_cntr = 0
        init_list = [0] * perm_length #make the first item i.e. [0,0,0,0] if using 4 groups
        perms = [init_list] #add it as the first list in the array
        while pos_cntr < perm_length: #pos_cntr will be the place edit i.e. [x,0,0,0] and length if the length of the array, i.e. 4
            cnt = 1 #has to be 1 not 0 for < perm_max[x] to work
            temp_array = perms.copy() #copy the array as it stands to a new one for it to be edited before being added back
            while cnt < perm_max[pos_cntr]: #counter has to be less than the max. available groups (i.e max. permutations)
                for list in temp_array: #for each list in the array change a value and add it to the permutations array
                    temp_list = list.copy()
                    temp_list[pos_cntr] = cnt
                    perms.append(temp_list)
                cnt += 1
            pos_cntr += 1
    #NEED TO ADD ABILITY TO PARSE MORE THAN ONE WILDCARD!
    ligand_num = 1
    for perm in perms: #for each permutation in the permutations array
        mol = Chem.MolFromSmiles(s_list[perm[0]]) #mol = the nth (the value of [x,0,0,0] in the array (perm[0]) item in the s_list
        pos = 1 #set for the first nt_grp
        while pos <= NTGps_to_use: #if ntgrp to use = 1 then will only pass once, if 0 it won't pass at all
            fg = Chem.MolFromSmiles(nt_list[perm[pos]]) #get the correct group from the list
            mol = replace_R(mol, fg) #replace the position of the wildcard with the new group
            pos += 1 #repeat
        while pos < perm_length:
            fg = Chem.MolFromSmiles(t_list[perm[pos]]) #same as above but with the t groups
            mol = replace_R(mol, fg)
            pos += 1
        # mol.SetProp("Ligand_Number", ligand_num)
        Chem.AddHs(mol)
        AllChem.EmbedMolecule(mol)
        AllChem.UFFOptimizeMolecule(mol, maxIters=500)
        sdf_file = os.path.join('data', name, name + '.sdf')
        sdf_open = open(sdf_file, 'a') #open in append mode, saves holding 1000s of compounds in memory
        writeMol(mol, sdf_open) #write them to the sdf
        print('Generated ligand ' + str(ligand_num) + '.')
        ligand_num += 1
    print("Generation complete. Molecules saved in " + sdf_file)



import os
import csv
import rdkit
from rdkit import Chem
import gendock.scripts.molgen as mg
import gendock.scripts.docking as dk
import gendock as gd

def chkDirectory():
    homeDir = os.getcwd()
    if not os.path.exists('receptors'):
        print("No receptors directory found in the current working directory, quitting GenDock.")
        exit()
    else:
        return

def generate(name, target_mass, *args, **kwargs):
    data_path = os.path.join('data', name)
    if not os.path.exists(data_path):
        os.mkdir(data_path)
    nligands = kwargs.get('nligands', 0)
    mol = kwargs.get('mol', None)
    if mol is not None:
        if not type(mol) == rdkit.Chem.rdchem.Mol:
            print("mol is not an RDKit mol type.")
            return
    mg.molgenerate(name, target_mass, nligands, mol, gd.s_list, gd.nt_list, gd.t_list)

def dock(name, *args, **kwargs):
    ligand_num = kwargs.get('ligand_num', 1)
    receptors = kwargs.get('receptors', None)
    rnum = len(receptors)
    if rnum == 0:
        print('No receptor names given, quitting.')
        return
    # check if receptors exist
    for r in receptors:
        rpath = os.path.join('receptors', r + ".pdbqt")
        print(rpath)
        if not os.path.exists(rpath):
            return
    # prepare csv for results
    if ligand_num == 1:
        headers = ["Ligand", "Mw", "LogP", "SMILES", "NSC"]
        csv_fname = str(name) + "_results.csv"
        csv_file = os.path.join('data', name, csv_fname)
        for receptor in receptors:
            r_header = receptor + " Binding Affinity (kcal/mol)"
            headers.append(r_header)
        if not os.path.exists(os.path.join('data', name)):
            os.mkdir(os.path.join('data', name))
        with open(csv_file, "a") as f:
            writer = csv.writer(f, delimiter=",")
            writer.writerow(headers)
    # if not os.path.exists(receptor_pdbqt):
    #     print("Preparing receptors...")
    #     convertMol(receptor_pdb, receptor_pdbqt)
    # else:
    #     print("Receptor PDBQT found.")
    # convertMol(receptor_pdb, receptor_pdbqt)
    #dock
    dk.moldocking(name, ligand_num, receptors)





from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()


setup(
    name='gendock',
    version='1.0',
    description='Python package for generating and docking molecules against macromolecules',
    url='https://github.com/philspence/gendock',
    author='Phil Spence',
    author_email='philspence91@gmail.com',
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.7',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux'
    ]
)
